"use strict";
const ts = require("typescript");
const path = require("path");
const babel = require("babel-core");
const fs_extra_p_1 = require("fs-extra-p");
const bluebird_1 = require("bluebird");
const declarationGenerator_1 = require("./declarationGenerator");
const markdown = require("markdown-it");
const __awaiter = require("./awaiter");
const basePath = process.argv.length === 3 ? path.resolve(process.argv[2]) : process.cwd();
const tsConfigPath = path.join(basePath, "tsconfig.json");
main()
    .catch(error => {
    if (error instanceof CompilationError) {
        for (let diagnostic of error.errors) {
            if (diagnostic.file == null) {
                console.log(diagnostic.messageText);
                continue;
            }
            const location = diagnostic.file.getLineAndCharacterOfPosition(diagnostic.start);
            const message = ts.flattenDiagnosticMessageText(diagnostic.messageText, '\n');
            console.log(`${diagnostic.file.fileName} (${location.line + 1},${location.character + 1}): ${message}`);
        }
    }
    else {
        console.error(error.stack || error.message || error);
    }
    process.exit(-1);
});
function main() {
    return __awaiter(this, void 0, void 0, function* () {
        const jsonResult = ts.parseConfigFileTextToJson(tsConfigPath, yield fs_extra_p_1.readFile(tsConfigPath, "utf8"));
        if (jsonResult.error != null) {
            throw new CompilationError([jsonResult.error]);
        }
        const result = ts.parseJsonConfigFileContent(jsonResult.config, {
            readDirectory: ts.sys.readDirectory
        }, basePath);
        checkErrors(result.errors);
        yield compile(result, jsonResult.config);
    });
}
function compile(config, tsConfig) {
    return __awaiter(this, void 0, void 0, function* () {
        const compilerOptions = config.options;
        const declarationConfig = tsConfig.declaration;
        if (declarationConfig != null) {
            compilerOptions.declaration = true;
        }
        compilerOptions.noEmitOnError = true;
        const program = ts.createProgram(config.fileNames, compilerOptions);
        checkErrors(ts.getPreEmitDiagnostics(program));
        const outDir = compilerOptions.outDir;
        if (outDir == null) {
            throw new Error("outDir is not specified in the compilerOptions");
        }
        yield fs_extra_p_1.ensureDir(outDir);
        const target = compilerOptions.target || ts.ScriptTarget.Latest;
        const fileToSourceMap = {};
        const promises = [];
        const emittedFiles = new Set();
        const declarationFiles = Array();
        const emitResult = program.emit(undefined, (fileName, data) => {
            emittedFiles.add(fileName);
            if (fileName.endsWith(".js")) {
                const sourceMapFileName = fileName + ".map";
                processCompiled(data, fileToSourceMap[sourceMapFileName], fileName, sourceMapFileName, promises);
            }
            else if (fileName.endsWith(".js.map")) {
                fileToSourceMap[fileName] = data;
            }
            else if (declarationConfig != null) {
                declarationFiles.push(ts.createSourceFile(fileName, data, target, true));
            }
        });
        if (tsConfig.docs != null) {
            const docs = generateDocs(program);
            if (docs.length !== 0) {
                writeDocFile(path.resolve(basePath, tsConfig.docs), docs);
            }
        }
        checkErrors(emitResult.diagnostics);
        if (emitResult.emitSkipped) {
            throw new Error("Emit skipped");
        }
        if (declarationFiles.length > 0) {
            for (let moduleName of Object.keys(declarationConfig)) {
                promises.push(declarationGenerator_1.generateDeclarationFile(moduleName, declarationFiles, compilerOptions, path.join(basePath, declarationConfig[moduleName]), basePath));
            }
        }
        yield bluebird_1.Promise.all(promises);
        promises.length = 0;
        yield removedOld(outDir, emittedFiles, promises);
        yield bluebird_1.Promise.all(promises);
    });
}
function writeDocFile(docOutFile, content) {
    return __awaiter(this, void 0, void 0, function* () {
        let existingContent = null;
        try {
            existingContent = yield fs_extra_p_1.readFile(docOutFile, "utf8");
        }
        catch (e) {
        }
        if (existingContent == null) {
            return fs_extra_p_1.outputFile(docOutFile, content);
        }
        else {
            const startMarker = "<!-- do not edit. start of generated block -->";
            const endMarker = "<!-- end of generated block -->";
            const start = existingContent.indexOf(startMarker);
            const end = existingContent.indexOf(endMarker);
            if (start != -1 && end != -1) {
                return fs_extra_p_1.outputFile(docOutFile, existingContent.substring(0, start + startMarker.length) + "\n" + content + "\n" + existingContent.substring(end));
            }
        }
        console.log("Write doc to " + docOutFile);
    });
}
function generateDocs(program) {
    const topicToProperties = new Map();
    for (let sourceFile of program.getSourceFiles()) {
        if (!sourceFile.isDeclarationFile) {
            for (let statement of sourceFile.statements) {
                if (statement.kind === ts.SyntaxKind.InterfaceDeclaration) {
                    const topicName = getComment(statement);
                    if (topicName == null) {
                        continue;
                    }
                    const interfaceDeclaration = statement;
                    const interfaceName = interfaceDeclaration.name.text;
                    const interfaceDescriptor = {
                        interfaceName: interfaceName,
                        heading: topicName,
                    };
                    let nameToProperty = topicToProperties.get(interfaceDescriptor);
                    for (let member of interfaceDeclaration.members) {
                        if (member.kind === ts.SyntaxKind.PropertySignature) {
                            const comment = getComment(member);
                            if (comment != null) {
                                if (nameToProperty == null) {
                                    nameToProperty = new Map();
                                    topicToProperties.set(interfaceDescriptor, nameToProperty);
                                }
                                nameToProperty.set(member.name.text, new PropertyDescriptor(interfaceName, comment));
                            }
                        }
                    }
                }
            }
        }
        function getComment(node) {
            const leadingCommentRanges = ts.getLeadingCommentRanges(sourceFile.text, node.pos);
            if (leadingCommentRanges == null || leadingCommentRanges.length === 0) {
                return null;
            }
            else {
                const commentRange = leadingCommentRanges[0];
                return sourceFile.text.slice(commentRange.pos + "/**".length, commentRange.end - "*/".length).trim();
            }
        }
    }
    return renderDocs(topicToProperties);
}
function renderDocs(topicToProperties) {
    let result = "";
    const md = markdown({
        typographer: true,
    })
        .disable(["link", "emphasis"]);
    topicToProperties.forEach((nameToProperty, interfaceDescriptor) => {
        result += `${anchor(interfaceDescriptor.interfaceName)}\n${interfaceDescriptor.heading}\n`;
        result += "| Name | Description\n";
        result += "| --- | ---";
        nameToProperty.forEach((descriptor, propertyName) => {
            result += `\n| ${anchor(descriptor.interfaceName + "-" + propertyName)}${propertyName} | `;
            const src = descriptor.description;
            result += src.includes("\n") ? md.render(src).trim().replace(/\n/g, " ") : src;
        });
        result += "\n";
    });
    return result;
}
function anchor(link) {
    return `<a class="anchor" id="user-content-${link}" href="#${link}" aria-hidden="true"></a>`;
}
class InterfaceDescriptor {
}
exports.InterfaceDescriptor = InterfaceDescriptor;
class PropertyDescriptor {
    constructor(interfaceName, description) {
        this.interfaceName = interfaceName;
        this.description = description;
    }
}
exports.PropertyDescriptor = PropertyDescriptor;
function removedOld(outDir, emittedFiles, promises) {
    return __awaiter(this, void 0, void 0, function* () {
        const files = yield fs_extra_p_1.readdir(outDir);
        for (let file of files) {
            if (file[0] !== "." && !file.endsWith(".d.ts")) {
                const fullPath = outDir + '/' + file;
                if (!file.includes(".")) {
                    removedOld(fullPath, emittedFiles, promises);
                    continue;
                }
                if (!emittedFiles.has(fullPath)) {
                    promises.push(fs_extra_p_1.unlink(fullPath));
                }
            }
        }
    });
}
function processCompiled(code, sourceMap, jsFileName, sourceMapFileName, promises) {
    const result = babel.transform(code, {
        inputSourceMap: JSON.parse(sourceMap),
        sourceMaps: true,
        filename: jsFileName,
    });
    promises.push(fs_extra_p_1.outputFile(jsFileName, result.code), fs_extra_p_1.outputJson(sourceMapFileName, result.map));
}
function checkErrors(errors) {
    if (errors.length !== 0) {
        throw new CompilationError(errors);
    }
}
class CompilationError extends Error {
    constructor(errors) {
        super("Compilation error");
        this.errors = errors;
    }
}
//# sourceMappingURL=builder.js.map